/* Copyright (C) 2016 Max Riechelmann <max.riechelmann@student.kit.edu>
   (Karlsruhe Institute of Technology)
   This library is free software; you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as published by the
   Free Software Foundation; either version 2.1 of the License, or (at your
   option) any later version.
   This library is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
   details.
   You should have received a copy of the GNU Lesser General Public License along
   with this library; if not, write to the Free Software Foundation, Inc., 51
   Franklin St, Fifth Floor, Boston, MA 02110, USA
   */


#include "myco-ipc.h"


int myco_error(char *error_message, int message_queue_id) {
    fprintf(stderr, "FATAL ERROR in ");
    fprintf(stderr, error_message);
    fprintf(stderr, " => %s\n", strerror(errno));
    if (errno != EIDRM) {
        fprintf(stderr, "Trying to remove message queue %d\n", message_queue_id);
        myco_remove_message_queue(message_queue_id);
    }
    exit(1);
}

int myco_create_private_message_queue() {
    int message_queue_id;

    message_queue_id = msgget(IPC_PRIVATE, PERM | IPC_CREAT);

    if (message_queue_id == -1) {
        fprintf(stderr, "FATAL ERROR in myco_create_private_message_queue() => %s\n", strerror(errno));
        error(1);
    }

    return message_queue_id;
}

int myco_create_global_message_queue() {
    int message_queue_id;

    message_queue_id = msgget(KEY, PERM | IPC_CREAT);
    if (message_queue_id == -1) {
        fprintf(stderr, "FATAL ERROR in myco_create_global_message_queue() => %s\n", strerror(errno));
        error(1);
    }

    return message_queue_id;
}

int myco_get_global_message_queue() {
    int message_queue_id;

    message_queue_id = msgget(KEY, 0);
    if (message_queue_id == -1) {
        fprintf(stderr, "FATAL ERROR in myco_get_global_message_queue() => %s\n", strerror(errno));
        error(1);
    }

    return message_queue_id;
}

int myco_remove_message_queue(int message_queue_id) {
    if (msgctl(message_queue_id, IPC_RMID, 0) == -1) {
        fprintf(stderr, "FATAL ERROR in myco_remove_message_queue()\n");
        exit(1);
    }

    return 0;
}

int myco_send(int message_queue_id, message msg) {
    message_item m;
    m.mtype = 1;

    memcpy(m.mtext, &msg, sizeof(msg));

    if (msgsnd(message_queue_id, &m, sizeof(m.mtext), 0) == -1) {
        fprintf(stderr, "FATAL ERROR (msgsnd): %s\n", strerror(errno));
    }

    return 0;
}

message myco_receive(int message_queue_id) {
    message msg = {0};
    message_item m;

    if ((msgrcv(message_queue_id, &m, sizeof(m.mtext), 0, 0)) < 0) {
        fprintf(stderr, "FATAL ERROR (msgrcv): %s %d\n", strerror(errno), sizeof(msg));
    }

    memcpy(&msg, m.mtext, sizeof(msg));

    return msg;
}

message myco_send_and_receive(message msg, int send_message_queue_id, int receive_message_queue_id) {
    myco_send(send_message_queue_id, msg);

    msg = myco_receive(receive_message_queue_id);

    return msg;
}

int myco_read_memory(pid_t source_pid, void *source_pointer, int source_length, void *target_pointer, int target_length) {
    struct iovec local[1];
    struct iovec remote[1];
    ssize_t nread;

    pid_t pid = source_pid;
    local[0].iov_base = target_pointer;
    local[0].iov_len = target_length;
    remote[0].iov_base = source_pointer;
    remote[0].iov_len = source_length;

    nread = process_vm_readv(pid, local, 1, remote, 1, 0);
    if (nread != target_length) {
        return -1;
    }

    return 0;
}

int myco_write_memory(pid_t target_pid, void *source_pointer, int source_length, void *target_pointer, int target_length) {
    struct iovec local[1];
    struct iovec remote[1];
    ssize_t nread;

    pid_t pid = target_pid;
    local[0].iov_base = target_pointer;
    local[0].iov_len = target_length;
    remote[0].iov_base = source_pointer;
    remote[0].iov_len = source_length;

    nread = process_vm_writev(pid, local, 1, remote, 1, 0);
    if (nread != target_length) {
        return -1;
    }

    return 0;
}
